(function (GCN) {

	'use strict';

	/**
	 * @class
	 * @name TagContainerAPI
	 */
	GCN.TagContainerAPI = GCN.defineChainback({
		/** @lends TagContainerAPI */

		/**
		 * @private
		 * @type {object<number, string>} Hash, mapping tag ids to their
		 *                                corresponding names.
		 */
		_tagIdToNameMap: null,

		/**
		 * @private
		 * @type {object<number, string>} Hash, mapping tag ids to their
		 *                                corresponding names for newly created
		 *                                tags.
		 */
		_createdTagIdToNameMap: {},

		/**
		 * @private
		 * @type {Array.<object>} A set of blocks that are are to be removed
		 *                        from this content object when saving it.
		 *                        This array is populated during the save
		 *                        process.  It get filled just before
		 *                        persisting the data to the server, and gets
		 *                        emptied as soon as the save operation
		 *                        succeeds.
		 */
		_deletedBlocks: [],

		/**
		 * @private
		 * @type {Array.<object>} A set of tags that are are to be removed from
		 *                        from this content object when it is saved.
		 */
		_deletedTags: [],

		/**
		 * Searching for a tag of a given id from the object structure that is
		 * returned by the REST API would require O(N) time.  This function,
		 * builds a hash that maps the tag id with its corresponding name, so
		 * that it can be mapped in O(1) time instead.
		 *
		 * @private
		 * @return {object<number,string>} A hash map where the key is the tag
		 *                                 id, and the value is the tag name.
		 */
		'!_mapTagIdsToNames': function () {
			var name;
			var map = {};
			var tags = this._data.tags;

			for (name in tags) {
				if (tags.hasOwnProperty(name)) {
					map[tags[name].id] = name;
				}
			}

			return map;
		},

		/**
		 * Retrieves data for a tag from the internal data object.
		 *
		 * @private
		 * @param {string} name The name of the tag.
		 * @return {!object} The tag data, or null if a there if no tag
		 *                   matching the given name.
		 */
		'!_getTagData': function (name) {
			return (this._data.tags && this._data.tags[name]) ||
			       (this._shadow.tags && this._shadow.tags[name]);
		},

		/**
		 * Get the tag whose id is `id'.
		 * Builds the `_tagIdToNameMap' hash map if it doesn't already exist.
		 *
		 * @todo: Should we deprecate this?
		 * @private
		 * @param {number} id Id of tag to retrieve.
		 * @return {object} The tag's data.
		 */
		'!_getTagDataById': function (id) {
			if (!this._tagIdToNameMap) {
				this._tagIdToNameMap = this._mapTagIdsToNames();
			}

			return this._getTagData(this._tagIdToNameMap[id] ||
				 this._createdTagIdToNameMap[id]);
		},

		/**
		 * Extracts the editables and blocks that have been rendered.
		 *
		 * @param {object} data The response object received from the
		 *                      renderTemplate() call.
		 * @return {object} An object containing two properties: an array of
		 *                  blocks, and an array of editables.
		 */
		'!_processRenderedTags': function (data) {
			var tags = this._getEditablesAndBlocks(data);
			this._storeRenderedEditables(tags.editables);
			this._storeRenderedBlocks(tags.blocks);
			return tags;
		},

		/**
		 * Get this content object's node.
		 *
		 * @public
		 * @function
		 * @name node
		 * @memberOf ContentObjectAPI
		 * @param {funtion(NodeAPI)=} success Optional callback to receive a
		 *                                    {@link NodeAPI} object as the
		 *                                    only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 * @return {NodeAPI} This object's node.
		 */
		'!node': function (success, error) {
			return this.folder().node();
		},

		/**
		 * Get this content object's parent folder.
		 *
		 * @public
		 * @function
		 * @name folder
		 * @memberOf ContentObjectAPI
		 * @param {funtion(FolderAPI)=} success Optional callback to receive a
		 *                                      {@link FolderAPI} object as the
		 *                                      only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 * @return {FolderAPI} This object's parent folder.
		 */
		'!folder': function (success, error) {
			var id = this._fetched ? this.prop('folderId') : null;
			return this._continue(GCN.FolderAPI, id, success, error);
		},

		/**
		 * Gets a tag of the specified id, contained in this content object.
		 *
		 * @name tag
		 * @function
		 * @memberOf TagContainerAPI
		 * @param {function} success
		 * @param {function} error
		 * @return TagAPI
		 */
		'!tag': function (id, success, error) {
			return this._continue(GCN.TagAPI, id, success, error);
		},

		/**
		 * Retrieves a collection of tags from this content object.
		 *
		 * @name tags
		 * @function
		 * @memberOf TagContainerAPI
		 * @param {object|string|number} settings (Optional)
		 * @param {function} success callback
		 * @param {function} error (Optional)
		 * @return TagContainerAPI
		 */
		'!tags': function () {
			var args = Array.prototype.slice.call(arguments);

			if (args.length === 0) {
				return;
			}

			var i;
			var j = args.length;
			var filter = {};
			var filters;
			var hasFilter = false;
			var success;
			var error;

			// Determine `success', `error', `filter'
			for (i = 0; i < j; ++i) {
				switch (jQuery.type(args[i])) {
				case 'function':
					if (success) {
						error = args[i];
					} else {
						success = args[i];
					}
					break;
				case 'number':
				case 'string':
					filters = [args[i]];
					break;
				case 'array':
					filters = args[i];
					break;
				default:
					return;
				}
			}

			if (jQuery.type(filters) === 'array') {
				var k = filters.length;
				while (k) {
					filter[filters[--k]] = true;
				}
				hasFilter = true;
			}

			var that = this;

			if (success) {
				this._read(function () {
					var tags = that._data.tags;
					var tag;
					var list = [];

					for (tag in tags) {
						if (tags.hasOwnProperty(tag)) {
							if (!hasFilter || filter[tag]) {
								list.push(that._continue(GCN.TagAPI, tags[tag],
									null, error));
							}
						}
					}

					success(list);
				}, error);
			}
		},

		/**
		 * Internal method to create a tag of a given tagtype in this content
		 * object.
		 *
		 * @private
		 * @param {string|number} construct The name of the construct on which
		 *                                  the tag to be created should be
		 *                                  derived from.  Or the id of that
		 *                                  construct.
		 * @param {string=} magicValue Optional property that will override the
		 *                             default values of this tag type.
		 * @param {function(TagAPI)=} success Optional callback that will
		 *                                    receive the newly created tag as
		 *                                    its only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 * @return {TagAPI} The newly created tag.
		 */
		'!_createTag': function () {
			var args = Array.prototype.slice.call(arguments);

			if (args.length === 0) {
				GCN.error('INVALID_ARGUMENTS', '`createTag()\' requires at ' +
					'least one argument.  See documentation.');

				return;
			}

			var success;
			var error;
			var magicValue;
			var construct = args[0];
			var i;
			var j = args.length;

			// Determine `success', `error', and `magicValue'.
			for (i = 1; i < j; ++i) {
				switch (jQuery.type(args[i])) {
				case 'string':
					magicValue = args[i];
					break;
				case 'function':
					if (success) {
						error = args[i];
					} else {
						success = args[i];
					}
					break;
				}
			}

			var that = this;

			// We use a uniqueId to avoid a fetus being created.
			// This is to avoid the following szenario:
			//
			// var tag1 = container.createTag(...);
			// var tag2 = container.createTag(...);
			// tag1 === tag2 // is true which is wrong
			//
			// However, for all other cases, where we get an existing object, we want this behaviour:
			// var folder1 = page(1).folder(...);
			// var folder2 = page(1).folder(...);
			// folder1 === folder2 // is true which is correct
			//
			// So, createTag() is different from other chainback methods
			// in that each invokation must create a new instance, while
			// other chainback methods must return the same.
			//
			// The id will be reset as soon as the tag object is
			// realized. This happens below as soon as we get a success
			// response with the correct tag id.
			var newId = GCN.uniqueId('TagApi-unique-');

			// First create a new TagAPI instance that will have this container
			// as its ancestor.  Also acquire a lock on the newly created tag
			// object so that any further operations on it will be queued until
			// we release the lock.
			var tag = this._continue(GCN.TagAPI, newId)._procure();

			this.node().constructs(function (constructs) {
				var constructId;

				if ('number' === jQuery.type(construct)) {
					constructId = construct;
				} else if (constructs[construct]) {
					constructId = constructs[construct].constructId;
				} else {
					var err = GCN.createError('CONSTRUCT_NOT_FOUND',
						'Cannot find constuct `' + construct + '\'',
						constructs);

					GCN.handleError(err, error);

					return;
				}

				var restUrl = GCN.settings.BACKEND_PATH + '/rest/'
				            + that._type + '/newtag/' + that._data.id + '?'
				            + jQuery.param({constructId: constructId});

				that._authAjax({
					type  : 'POST',
					url   : restUrl,
					json  : {magicValue: magicValue},
					error : function (xhr, status, msg) {
						GCN.handleHttpError(xhr, msg, error);
						tag._vacate();
					},
					success: function (response) {
						if (GCN.getResponseCode(response) === 'OK') {
							var data = response.tag;

							tag._name    = data.name;
							tag._data    = data;
							tag._fetched = true;

							// The tag's id is still newId from
							// above. We have to realize the tag so that
							// it gets the correct id. The new id
							// changes its hash, so it must also be
							// removed and reinserted from the caches.
							tag._removeFromTempCache();
							tag._setHash(data.id)._addToCache();

							// Add this tag into the tag's container `_shadow'
							// object, and `_tagIdToNameMap hash'.

							var shouldCreateObjectIfUndefined = true;

							// Add this tag into the `_shadow' object.
							that._update('tags.' + data.name, data, error,
								shouldCreateObjectIfUndefined);

							// TODO: We need to store the tag inside the
							// `_data' object for now.  A change should be made
							// so that when containers are saved, the data in
							// the _shadow object is properly transfered into
							// the _data object.

							that._data.tags[data.name] = data;

							if (!that.hasOwnProperty('_createdTagIdToNameMap')) {
								that._createdTagIdToNameMap = {};
							}

							that._createdTagIdToNameMap[data.id] = data.name;

							if (success) {
								success(tag);
							}
						} else {
							tag._die(GCN.getResponseCode(response));
							GCN.handleResponseError(response, error);
						}

						// Hold onto the mutex until this tag object has been
						// fully realized and placed inside its container.
						tag._vacate();
					}
				});
			}, error);

			return tag;
		},

		/**
		 * Internal method to delete the specified tag from this content
		 * object.
		 *
		 * @private
		 * @param {string} id The id of the tag to be deleted.
		 * @param {function(TagContainerAPI)=} success Optional callback that
		 *                                             receive this object as
		 *                                             its only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		'!_removeTag': function (id, success, error) {
			this.tag(id).remove(success, error);
		},

		/**
		 * Internal method to delete a set of tags from this content object.
		 *
		 * @private
		 * @param {Array.<string>} ids The ids of the set of tags to be
		 *                             deleted.
		 * @param {function(TagContainerAPI)=} success Optional callback that
		 *                                             receive this object as
		 *                                             its only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		'!_removeTags': function (ids, success, error) {
			var that = this;

			this.tags(ids, function (tags) {
				var j = tags.length;

				while (j) {
					tags[--j].remove(null, error);
				}

				if (success) {
					that.save(success, error);
				}
			}, error);
		},

		/**
		 * Given a data object received from a "/rest/page/render" call, map
		 * the blocks and editables into a list of each.
		 *
		 * Note that if a tag is both an editable and a block, it will be
		 * listed in both the blocks list and in the editables list.
		 *
		 * @param {object} data
		 * @return {object<string, Array.<object>>} A map containing a set of
		 *                                          editables and a set of
		 *                                          blocks.
		 */
		'!_getEditablesAndBlocks': function (data) {
			if (!data || !data.tags) {
				return {
					blocks: [],
					editables: []
				};
			}

			var tag;
			var tags = data.tags;
			var numTags = tags.length;
			var numEditables;
			var blocks = [];
			var editables = [];
			var isBlock;
			var i;
			var j;

			for (i = 0; i < numTags; i++) {
				tag = tags[i];

				if (tag.editables) {
					numEditables = tag.editables.length;
					for (j = 0; j < numEditables; j++) {
						tag.editables[j].tagname = tag.tagname;
					}
					editables = editables.concat(tag.editables);
				}

				/*
				 * Depending on tag.onlyeditables to determine whether or not a
				 * given tag is a block or editable is unreliable since it is
				 * possible to have a block which only contains editables:
				 *
				 * {
				 *  "tagname":"content",
				 *  "editables":[{
				 *	  "element":"GENTICS_EDITABLE_1234",
				 *	  "readonly":false,
				 *	  "partname":"editablepart"
				 *  }],
				 *  "element":"GENTICS_BLOCK_1234",
				 *  "onlyeditables":true
				 *  "tagname":"tagwitheditable"
				 * }
				 *
				 * in the above example, tag.onlyeditable is true but tag is a
				 * block, since the tag's element and the editable's element
				 * are not the same.
				 */

				if (!tag.editables || tag.editables.length > 1) {
					isBlock = true;
				} else {
					isBlock = (1 === tag.editables.length)
						&& (tag.editables[0].element !== tag.element);
				}

				if (isBlock) {
					blocks.push(tag);
				}
			}

			return {
				blocks: blocks,
				editables: editables
			};
		}

	});

}(GCN));
